"""Julian day

Julian day is the interval of time in days since January 1, 4713 BC 
Greenwich noon. It is convenient for calculating intervals between two 
historical dates.

References:
* Julian day - Wikipedia
  http://en.wikipedia.org/wiki/Julian_day

Copyright (c) 2012 Masaaki Shibata <mshibata@emptypage.jp>

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the 
"Software"), to deal in the Software without restriction, including 
without limitation the rights to use, copy, modify, merge, publish, 
distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to 
the following conditions:

The above copyright notice and this permission notice shall be 
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY 
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


__revision__ = '$Rev: 1974M $'


import sys
import time


JD_8_AD = 1722579.0
WARNING_EARLIER_THAN_8_AD = """
NOTE: Dates earlier than 8 AD in Julian calendar may be incorrect because 
the leap years had occurred not correctly in that time.
""".strip()


class JD(object):
    
    """Julian date object """
    
    offset = 0
    """Offset from the Julian day epoch (12:00 Jan 1, 4713 BC GMT) in days
    
    Subclasses may override this value to get proper jd() value and 
    comparison result with other JD objects.
    """
    
    def __init__(self, value):
        
        """Create Julian date object with given value
        
        Value may be int, float or other JD object. If it is JD 
        object, new instance will be initialized with the value.jd().
        
        >>> jd1 = JD(2455972)       # init with int
        >>> jd1
        JD(2455972)
        >>> jd2 = JD(2455972.03851) # with float
        >>> jd2
        JD(2455972.03851)
        >>> jd3 = JD(jd2)           # and with other JD object
        >>> jd3
        JD(2455972.03851)
        """
        
        if isinstance(value, JD):
            value = value.jd() + self.offset
        self.value = value
    
    def __str__(self):
        return '%s %s' % (self.__class__.__name__, self.value)
    
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.value)
    
    def __cmp__(self, other):
        
        """JD objects having same values are treated as equal
        
        >>> jd1 = JD(2455972.03851)
        >>> jd2 = JD(2455972.03851)
        >>> jd3 = JD(2455972)
        >>> jd1 == jd2
        True
        >>> jd1 == jd3
        False
        """
        
        if isinstance(other, JD):
            return cmp(self.jd(), other.jd())
    
    def __hash__(self):
        
        """JD objects are hashable. """
        
        return hash(self.jd())
    
    def __int__(self):
        return int(self.value)
    
    def __long__(self):
        return long(self.value)
    
    def __float__(self):
        return float(self.value)
    
    def __add__(self, other):
        
        """Return new JD object having its result
        
        i.e. JD objects are immutable.
        
        >>> JD(2455106) + 1
        JD(2455107)
        """
        
        return self.__class__(self.value + other)
    
    def __sub__(self, other):
        
        """Return new JD object having its result (subtracting int/float) 
        or duration value in days (subtracting other JD object)
        
        >>> JD(2455106) - 1
        JD(2455105)
        >>> JD(2455106) - JD(2455006)
        100
        """
        
        if isinstance(other, JD):
            return self.jd() - other.jd()
        else:
            return self.__class__(self.value - other)
    
    def __mod__(self, other):
        
        """Return object's value % other (remainder)
        
        >>> JD(2455106) % 5
        1
        """
        
        return self.value % other
    
    def jd(self):
        
        """Return Julian day value of the instance in int/float
        
        This method is important for subclassed JD objects because they 
        are compared with each other by results of calling this method.
        """
        
        return self.value - self.offset
    
    def jdn(self):
        
        """Return JDN (Julian day number)
        
        Julian day number is the integer part of the JD.
        """
        
        return int(self.jd())
    
    def weekday(self):
        
        """Return day of week of the date (0-6, 0 for Sunday)
        
        >>> JD(2455105.5).weekday() # 2009-10-01 00:00:00
        4
        >>> JD(2455106).weekday()   # 2009-10-01 12:00:00
        4
        >>> JD(2455106.5).weekday() # 2009-10-02 00:00:00
        5
        """
        
        return int(self.jd() + 1.5) % 7
    
    def isoweekday(self):
        
        """Return day of week of the date (1-7, 1 for Sunday)
        
        >>> JD(2455105.5).isoweekday()  # 2009-10-01 00:00:00
        5
        >>> JD(2455106).isoweekday()    # 2009-10-01 12:00:00
        5
        >>> JD(2455106.5).isoweekday()  # 2009-10-02 00:00:00
        6
        """
        
        return self.weekday() + 1
    
    def _calendar_datetime(self, func_calendar_datetime_at_jd):
        
        return func_calendar_datetime_at_jd(self.jd())
    
    def julian_calendar_datetime(self):
        
        """Return (year, month, day, hour, minute, second) tuple in Julian 
        calendar date
        
        Please do not confuse Julian calendar with Julian day; they are 
        quite different ideas. Julian calendar is the calendar system used 
        until Oct 4, 1582 (JDN 2299160). The next day of the day is Oct 
        15, 1582 (JDN 2299161), the date in Gregorian calendar which is 
        widely used now.
        
        >>> JD(2299160).julian_calendar_datetime()
        (1582, 10, 4, 12, 0, 0)
        >>> JD(2299161).julian_calendar_datetime()
        (1582, 10, 5, 12, 0, 0)
        """
        
        return self._calendar_datetime(julian_calendar_datetime_at_jd)
        
    def gregorian_calendar_datetime(self):
        
        """Return (year, month, day, hour, minute, second) tuple in 
        Gregorian calendar date
        
        Gregorian calendar is the current calendar system which is widely 
        used. It starts from Oct 15, 1582 (JDN 2299161). The previous day 
        of the day is Oct 4, 1582 (JDN 2299160), the date in Julian 
        calendar.
        
        >>> JD(2299160).gregorian_calendar_datetime()
        (1582, 10, 14, 12, 0, 0)
        >>> JD(2299161).gregorian_calendar_datetime()
        (1582, 10, 15, 12, 0, 0)
        """
        
        return self._calendar_datetime(gregorian_calendar_datetime_at_jd)
    
    def calendar_datetime(self):
        
        """Return (year, month, day, hour, minute, second) tuple in 
        historically appropriate calendar date
        
        It returns date in Julian calendar if self.jd() < 2299160.5, 
        otherwise in Gregorian calendar.
        
        >>> JD(2299160).calendar_datetime()
        (1582, 10, 4, 12, 0, 0)
        >>> JD(2299161).calendar_datetime()
        (1582, 10, 15, 12, 0, 0)
        >>> JD(2454115.05486).calendar_datetime()
        (2007, 1, 14, 13, 18, 59)
        """
        
        return self._calendar_datetime(calendar_datetime_at_jd)
    
    @classmethod
    def _from_calendar_datetime(cls, func_jd_at_calendar_datetime,
                                year, month, day, hour=0, minute=0, second=0):
        
        jd = func_jd_at_calendar_datetime(year, month, day, hour, minute, second)
        return cls(jd + cls.offset)
    
    @classmethod
    def from_julian_calendar_datetime(cls, year, month, day, hour=0, minute=0, second=0):
        
        """Create JD object from the jurian calendar datetime
        
        >>> JD.from_julian_calendar_datetime(1582, 10, 4, 12, 0, 0)
        JD(2299160.0)
        """
        
        return cls._from_calendar_datetime(jd_at_julian_calendar_datetime,
                                           year, month, day, hour, minute, second)
    
    @classmethod
    def from_gregorian_calendar_datetime(cls, year, month, day, hour=0, minute=0, second=0):
        
        """Create JD object from the gregorian calendar datetime
        
        >>> JD.from_gregorian_calendar_datetime(2012, 2, 15)
        JD(2455972.5)
        """
        
        return cls._from_calendar_datetime(jd_at_gregorian_calendar_datetime,
                                           year, month, day, hour, minute, second)
    
    @classmethod
    def from_calendar_datetime(cls, year, month, day, hour=0, minute=0, second=0):
        
        """Create JD object from the calendar datetime
        
        >>> JD.from_calendar_datetime(2012, 2, 15)
        JD(2455972.5)
        """
        
        return cls._from_calendar_datetime(jd_at_calendar_datetime,
                                           year, month, day, hour, minute, second)
    
    @classmethod
    def now(cls):
        
        """Return current JD """
        
        tm = time.gmtime()
        n = jdn_at_gregorian_calendar_date(tm.tm_year, tm.tm_mon, tm.tm_mday)
        f = tm.tm_hour/24.0 + tm.tm_min/1440.0 + tm.tm_sec/86400.0
        return cls(n - 0.5 + f + cls.offset)


class MJD(JD):
    
    """Modified Julian date object
    
    The epoch of the MJD is 00:00 Nov 17, 1858 GMT; it starts from 
    midnight rather than noon.
    """
    
    offset = -2400000.5


def create_cjd_class(adjustment):
    
    """Create chronological Julian date class
    
    The epoch of the CJD is 00:00 Jan 1, 4713 BC GMT; it starts from 
    midnight rather than noon. Also, CJD treats (simple) timezone; this is 
    the local edition of the JD.
    
    >>> jd1 = JD(2455973.0) # 2012-02-15 12:00:00 GMT
    >>> JST_CJD = create_cjd_class(fractional_day(hours=9)) # +0900
    >>> cjd1 = JST_CJD(jd1)
    >>> cjd1
    CJD(2455972.875)
    >>> cjd1 == jd1
    True
    >>> cjd1.jd()
    2455973.0
    >>> cjd1.calendar_datetime()
    (2012, 2, 15, 21, 0, 0)
    >>> cjd2 = JST_CJD.from_calendar_datetime(2012, 2, 15, 21, 0, 0)
    >>> cjd2
    CJD(2455972.875)
    """
    
    _adjustment = adjustment    # to avoid namespace issue
    
    class CJD(JD):
        
        offset = -0.5 + _adjustment
        adjustment = _adjustment
        
        def _calendar_datetime(self, func_calendar_datetime_at_jd):
            
            return func_calendar_datetime_at_jd(self.jd() + _adjustment)
        
        @classmethod
        def _from_calendar_datetime(cls, func_jd_at_calendar_datetime,
                                    year, month, day, hour=0, minute=0, second=0):
            
            jd = func_jd_at_calendar_datetime(year, month, day, hour, minute, second)
            return cls(jd + cls.offset - _adjustment)
    
    return CJD


def fractional_day(days=0, hours=0, minutes=0, seconds=0):
    
    """Return fractional decimal day value of the time
    
    1 for 24 hours, 0.5 for 12 hours.
    
    >>> fractional_day()        # zero
    0.0
    >>> fractional_day(0, 6)    # 6 hours
    0.25
    >>> fractional_day(hours=12)
    0.5
    >>> fractional_day(0, 18, 0, 0)
    0.75
    >>> fractional_day(0, 10, 10, 30)
    0.4239583333333333
    """
    
    return days + (hours*3600 + minutes*60 + seconds) / 86400.0


def standard_time(fd):
    
    """Return tuple of (hour, minute, second) for the fractional day
    
    >>> standard_time(0.0)
    (0, 0, 0)
    >>> standard_time(0.25)
    (6, 0, 0)
    >>> standard_time(0.5)
    (12, 0, 0)
    >>> standard_time(0.75)
    (18, 0, 0)
    >>> standard_time(0.4239583333333333)
    (10, 10, 30)
    """
    
    total_seconds = fd * 86400.0
    hour, r = divmod(total_seconds, 3600.0)
    minute, second = divmod(r, 60.0)
    return int(hour), int(minute), int(second)


def jdn_at_julian_calendar_date(year, month, day):
    
    """Return the JDN of the Julian calendar date (at noon GMT)
    
    >>> jdn_at_julian_calendar_date(-4712, 1, 1) # the epoch
    0
    >>> jdn_at_julian_calendar_date(1582, 10, 4)
    2299160
    """
    
    a = (14 - month) / 12
    y = year + 4800 - a
    m = month + 12 * a - 3
    jdn = day + (153*m+2)/5 + y*365 + y/4 - 32083
    
    if jdn < JD_8_AD:
        print >>sys.stderr, WARNING_EARLIER_THAN_8_AD
    
    return jdn


def jdn_at_gregorian_calendar_date(year, month, day):
    
    """Return the JDN of the Gregorian calendar date (at noon GMT)
    
    >>> jdn_at_gregorian_calendar_date(2009, 10, 1)
    2455106
    >>> jdn_at_gregorian_calendar_date(2012, 2, 13)
    2455971
    """
    
    a = (14 - month) / 12
    y = year + 4800 - a
    m = month + 12 * a - 3
    return day + (153*m+2)/5 + y*365 + y/4 - y/100 + y/400 - 32045


def jdn_at_calendar_date(year, month, day):
    
    """Return the JDN value of the historical date (at noon GMT)
    
    >>> jdn_at_julian_calendar_date(-4712, 1, 1) # the epoc
    0
    >>> jdn_at_calendar_date(1582, 10, 4) # the last day of the Julian cal.
    2299160
    >>> jdn_at_calendar_date(1582, 10, 10)
    Traceback (most recent call last):
        ...
    ValueError: invalid calendar date
    >>> jdn_at_calendar_date(1582, 10, 15) # the first day of the Gregorian cal.
    2299161
    >>> jdn_at_calendar_date(2009, 10, 1)
    2455106
    """
    
    if year < 1582:
        return jdn_at_julian_calendar_date(year, month, day)
    if year > 1582:
        return jdn_at_gregorian_calendar_date(year, month, day)
    if month < 10:
        return jdn_at_julian_calendar_date(year, month, day)
    if month > 10:
        return jdn_at_gregorian_calendar_date(year, month, day)
    if day < 5:
        return jdn_at_julian_calendar_date(year, month, day)
    if day > 14:
        return jdn_at_gregorian_calendar_date(year, month, day)
    raise ValueError('invalid calendar date')


def julian_calendar_date_at_jdn(jdn):
    
    """Return 3-tuple of year, month and day in the Julian calendar date
    
    >>> julian_calendar_date_at_jdn(2299160) # the last day of the Julian calendar
    (1582, 10, 4)
    >>> julian_calendar_date_at_jdn(2299161) # the first day of the Gregorian calendar
    (1582, 10, 5)
    >>> julian_calendar_date_at_jdn(2455106)
    (2009, 9, 18)
    """
    
    if jdn < JD_8_AD:
        print >>sys.stderr, WARNING_EARLIER_THAN_8_AD
    
    b = 0
    c = jdn + 32082
    
    d = (4 * c + 3) / 1461
    e = c - (1461 * d) / 4
    m = (5 * e + 2) / 153
    
    day = e - (153 * m + 2) / 5 + 1
    month = m + 3 - 12 * (m / 10)
    year = b * 100 + d - 4800 + m / 10
    
    return year, month, day


def gregorian_calendar_date_at_jdn(jdn):
    
    """Return 3-tuple of year, month and day in the Gregorian calendar date
    
    >>> gregorian_calendar_date_at_jdn(2299160) # the last day of the Julian calendar
    (1582, 10, 14)
    >>> gregorian_calendar_date_at_jdn(2299161) # the first day of the Gregorian calendar
    (1582, 10, 15)
    >>> gregorian_calendar_date_at_jdn(2455106)
    (2009, 10, 1)
    """
    
    a = jdn + 32044
    b = (4 * a + 3) / 146097
    c = a - (b * 146097) / 4
    
    d = (4 * c + 3) / 1461
    e = c - (1461 * d) / 4
    m = (5 * e + 2) / 153
    
    day = e - (153 * m + 2) / 5 + 1
    month = m + 3 - 12 * (m / 10)
    year = b * 100 + d - 4800 + m / 10
    
    return year, month, day


def calendar_date_at_jdn(jdn):
    
    """Return 3-tuple of year, month and day in the calendar date
    
    If the JDN value is < 2299161, returned date will be in the Julian 
    calendar; otherwise, one in the Gregorian calendar.
    
    >>> calendar_date_at_jdn(2299160) # the last day of the Julian calendar
    (1582, 10, 4)
    >>> calendar_date_at_jdn(2299161) # the first day of the Gregorian calendar
    (1582, 10, 15)
    >>> calendar_date_at_jdn(2455106)
    (2009, 10, 1)
    """
    
    if jdn < 2299161:
        return julian_calendar_date_at_jdn(jdn)
    else:
        return gregorian_calendar_date_at_jdn(jdn)


def _jd_at_calendar_datetime(func_jdn_at_calendar_date,
                             year, month, day, hour=0, minute=0, second=0):
    
    jdn = func_jdn_at_calendar_date(year, month, day)
    offset = fractional_day(0, hour, minute, second) - 0.5
    return jdn + offset


def jd_at_julian_calendar_datetime(year, month, day, hour=0, minute=0, second=0):
    
    """Return Julian day value in float for the Julian dalendar datetime
    
    >>> jd_at_julian_calendar_datetime(-4712, 1, 1, 12, 0, 0)
    0.0
    >>> jd_at_julian_calendar_datetime(1582, 10, 4)
    2299159.5
    """
    
    return _jd_at_calendar_datetime(jdn_at_julian_calendar_date,
                                    year, month, day, hour, minute, second)


def jd_at_gregorian_calendar_datetime(year, month, day, hour=0, minute=0, second=0):
    
    """Return Julian day value in float for the Julian dalendar datetime
    
    >>> jd_at_gregorian_calendar_datetime(1582, 10, 15)
    2299160.5
    >>> jd_at_gregorian_calendar_datetime(2012, 2, 15, 12, 0, 0)
    2455973.0
    """
    
    return _jd_at_calendar_datetime(jdn_at_gregorian_calendar_date,
                                    year, month, day, hour, minute, second)


def jd_at_calendar_datetime(year, month, day, hour=0, minute=0, second=0):
    
    """Return Julian day value in float for the datetime
    
    >>> jd_at_calendar_datetime(-4712, 1, 1, 12, 0, 0)
    0.0
    >>> jd_at_calendar_datetime(1582, 10, 4)
    2299159.5
    >>> jd_at_calendar_datetime(1582, 10, 15)
    2299160.5
    >>> jd_at_calendar_datetime(2012, 2, 15)
    2455972.5
    >>> jd_at_calendar_datetime(2012, 2, 15, 12, 0, 0)
    2455973.0
    """
    
    return _jd_at_calendar_datetime(jdn_at_calendar_date,
                                    year, month, day, hour, minute, second)


def _calendar_datetime_at_jd(func_calendar_date_at_jdn, jd):
    
    datepart, timepart = divmod(jd-0.5, 1)
    year, month, day = func_calendar_date_at_jdn(int(datepart)+1)
    hour, minute, second = standard_time(timepart)
    return year, month, day, hour, minute, second


def julian_calendar_datetime_at_jd(jd):
    
    """Return (year, month, day, hour, minute, second) tuple in Julian 
    calendar for the Julian day
    
    >>> julian_calendar_datetime_at_jd(0)
    (-4712, 1, 1, 12, 0, 0)
    >>> julian_calendar_datetime_at_jd(2299159.5)
    (1582, 10, 4, 0, 0, 0)
    """
    
    return _calendar_datetime_at_jd(julian_calendar_date_at_jdn, jd)


def gregorian_calendar_datetime_at_jd(jd):
    
    """Return (year, month, day, hour, minute, second) tuple in Gregorian 
    calendar for the Julian day
    
    >>> gregorian_calendar_datetime_at_jd(2299160.5)
    (1582, 10, 15, 0, 0, 0)
    >>> gregorian_calendar_datetime_at_jd(2455973.0)
    (2012, 2, 15, 12, 0, 0)
    """
    
    return _calendar_datetime_at_jd(gregorian_calendar_date_at_jdn, jd)


def calendar_datetime_at_jd(jd):
    
    """Return (year, month, day, hour, minute, second) tuple in calendar 
    for the Julian day
    
    >>> calendar_datetime_at_jd(0)
    (-4712, 1, 1, 12, 0, 0)
    >>> calendar_datetime_at_jd(2299159.5)
    (1582, 10, 4, 0, 0, 0)
    >>> calendar_datetime_at_jd(2299160.5)
    (1582, 10, 15, 0, 0, 0)
    >>> calendar_datetime_at_jd(2455973.0)
    (2012, 2, 15, 12, 0, 0)
    >>> calendar_datetime_at_jd(2455972.5)
    (2012, 2, 15, 0, 0, 0)
    >>> calendar_datetime_at_jd(2455973.0)
    (2012, 2, 15, 12, 0, 0)
    """
    
    return _calendar_datetime_at_jd(calendar_date_at_jdn, jd)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
